from math import exp
from queue import Empty
import threading
from loguru import logger
from time import sleep
from DataObject import DataObject
from enum import Enum

from SlaveHandler import SlaveHandler


POISON_PILL = DataObject(name="PoisonPill", topic="kill", payload=None)
TIMEOUT_PILL = DataObject(name="TimeoutPill", topic="timeout", payload=None)

LOOP_SLICE = 0.1  # seconds
GRACE_PERIOD = 50

class e_State(Enum):
    IDLE = 0
    RUNNING = 1
    WAITING = 2

class RigCmdHandler(threading.Thread):
    def __init__(self, config, inQueue):
        super().__init__()

        self.config = config
        localConfig = self.config["master"]
        self.slaveNames = localConfig["slaves"].split(' ')
        self.slaves = []
        self.inQueue = inQueue

    def run(self):
        logger.debug("RigCmdHandler loop started")
        state = e_State.IDLE
        waitCnt = 0
        experiment = None

        while True:
            try:
                dataObject = self.inQueue.get(timeout=LOOP_SLICE)
            except Empty:
                dataObject = TIMEOUT_PILL
            if (dataObject == POISON_PILL):
                logger.error("RigCmdHandler swallowed the poison pill")
                break
            
            # logger.debug(f"State is {state}, WaitCnt is {waitCnt}")
            # Python 3.10 is required!!
            match state:
                case e_State.IDLE:
                    if (dataObject == TIMEOUT_PILL):
                        continue
                    if (dataObject.name == 'cmd') and (dataObject.payload["es"] == 'start'):
                        experiment = dataObject.payload["en"]
                        logger.info(f"start command received, switch to RUNNING state, experiment is {experiment}")
                        state = e_State.RUNNING
                        self.experimentStart(experiment)
                    elif (dataObject.name == 'cmd'):
                        logger.error(f"illegal command {dataObject.name} received in IDLE state")
                    else:
                        logger.error("illegal message received in IDLE state")
                case e_State.RUNNING:
                    if (dataObject == TIMEOUT_PILL):
                        continue
                    if (dataObject.name == 'cmd') and (dataObject.payload["es"] == 'stop'):
                        logger.info("stop command received, switch to WAITING state")
                        state = e_State.WAITING
                        waitCnt = 0
                    elif (dataObject.name == 'cmd'):
                        logger.error(f"illegal command {dataObject.name} received in RUNNING state")
                    else:
                        logger.error("illegal message received in RUNNING state")
                case e_State.WAITING:
                    if (dataObject == TIMEOUT_PILL):
                        waitCnt += 1
                        if (waitCnt >= GRACE_PERIOD):
                            logger.info("grace period is over, switch to IDLE state")
                            state = e_State.IDLE
                            self.experimentStop(experiment)
                    else:
                        logger.error("illegal message (not TIMEOUT) received in WAITING state")


    def experimentStart(self, experiment):
        logger.info(f"experiment started, {experiment}")
        for slaveName in self.slaveNames:
            slave = SlaveHandler(self.config, slaveName, experiment)
            slave.start()
            self.slaves.append(slave)

    def experimentStop(self, experiment):
        logger.info(f"experiment stopped, {experiment}")
        for slave in self.slaves:
            slave.stop()
        self.slaves.clear()


    def stop(self):
        self.inQueue.put(POISON_PILL)
        logger.debug("kill flag set")
